/*
 * NOISE.H  (imported from Peachey: Texturing And Modelling)
 *
 * cleaned-up version
 *
 *
 * We generate 3D noise by performing a lookup into a time-parametrized 
 * gradient noise table according to the "frozen turbulence"-theory.
 * This reduces per-vertex complexity and -more important- vertex-program
 * commands by 50%, enabling us to effectively perform the fBm perturbation
 * inside a vertex program.
 *
 * This requires generating a new gradient table as a per-frame setup which
 * can be either done in software or as a vertex state program. Since the 
 * vertex state program is itself limited to 128 commands, performing the 
 * setup on the host CPU should be given preference for tables of reasonable
 * sizes. This however does not impose a major restriction to speed, since
 * tables of 64 entries are fully sufficient and only two 1D-lookups, three
 * LERPs and one smoothstep are needed per entry.
 *
 */

#ifndef __NOISE_H__
#define __NOISE_H__

#define FLOOR(x)         ((int) (x) - ((x) < 0 && (x) != (int)(x)))
#define CEIL(x)          ((int) (x) + ((x) < 0 && (x) != (int)(x)))

#define TABSIZE          64
#define TABMASK          63
#define PERM(x)          GRADtab[4*((int)(x)&TABMASK)+3]
#define IND(ix,iy)       (int)(PERM((iy)+PERM((ix))))
#define RANDNR           ((float)rand()/(float)RAND_MAX)

#define LET(x,y)         (x)[0]=(y)[0];(x)[1]=(y)[1];(x)[2]=(y)[2]
#define LETS(p,x,y,z)	 (p)[0]=x;(p)[1]=y;(p)[2]=z
#define CLAMP(a,x,y)     ((a)<(x) ? (x) : (a)>(y) ? (y) : (a))
#define ABS(x)           ((x)<0 ? -(x) : (x))
#define VSMUL(x,scalar)  (x)[0]*=scalar;(x)[1]*=scalar;(x)[2]*=scalar
#define VADD(x,y)        (x)[0]+=(y)[0];(x)[1]+=(y)[1];(x)[2]+=(y)[2]

#define LERP(t,x0,x1)    ((1-t)*(x0)+(t)*(x1)) // Linear Interpolation
#define VLERP(res,t,x,y) (res)[0]=LERP((t)[0],(x)[0],(y)[0]);(res)[1]=LERP((t)[1],(x)[1],(y)[1]);(res)[2]=LERP((t)[2],(x)[2],(y)[2])

#define SMOOTHSTEP(x)    ((x)*(x)*(3-2*(x)))

#ifndef M_PI
#define M_PI 3.141592654f
#endif

static float gradientTab[TABSIZE*3];   // static gradient table, resides in host memory
static float GRADtab[TABSIZE*4];       // time parametrized gradient table, copied to program-vars

/*
 * INITIALIZES static gradient table
 *
 */ 
void gradientTabInit(int seed) {
    float z,r,theta;
    int i;
    srand(seed);
    for (i=0; i<TABSIZE; i++) {
        z=1.0f-2.0f*RANDNR;
        r=sqrt(1-z*z);
        theta=2*M_PI*RANDNR; // XXX: might lead to regular patterns!
        gradientTab[3*i  ]=r*cos(theta);
        gradientTab[3*i+1]=r*sin(theta);
        gradientTab[3*i+2]=0;
    }
}

/*
 * INITIALIZES permutation table
 *
 */
void permutationTabInit(int seed) {
	int i,j,tmp;
	srand(seed);
	// generate series of numbers
	for (i=0; i<TABSIZE; i++) {
		GRADtab[4*i+3]=i;
	}
	// permute them
	for (i=0; i<TABSIZE; i++) {
		j=(TABSIZE*rand())/RAND_MAX;		
		tmp=GRADtab[4*i+3];
		GRADtab[4*i+3]=GRADtab[4*j+3];
		GRADtab[4*j+3]=tmp;
	}
}

/*
 * GENERATE time-parametrized gradient table (acc. to "frozen turbulence"-theory)
 *
 */
void CalcGradient(float grad[3], float x) {
    int ix;
    float fx0,fx1;
    float wx[3];    
	float g0[3],g1[3];

    // Calculate lattice positions and interpolation coefficients
    ix=FLOOR(x);
    fx0=x-(float)ix;
    fx1=fx0-1.0f;
    wx[0]=SMOOTHSTEP(fx0);
	wx[1]=wx[0];
	wx[2]=wx[0];
	
    // Interpolate noise-triple
	LET(g0,&gradientTab[3*(ix&TABMASK)]);
	LET(g1,&gradientTab[3*((ix+1)&TABMASK)]);
    VLERP(grad,wx,g0,g1);
    normalize(grad);	
}

float lookupHeight(float x, float y) {
	if (x<0) x=-x;
	if (y<0) y=-y;
	int ix=FLOOR(x);
	int iy=FLOOR(y);
	float fx0=x-(float)ix;
	float fx1=fx0-1.0f;
	float fy0=y-(float)iy;
	float fy1=fy0-1.0f;	
	float n00,n01,n10,n11,nx0,nx1;
	float wx=SMOOTHSTEP(fx0);
	float wy=SMOOTHSTEP(fy0);
	int i;
	
	i=3*IND(ix,iy);
	n00=gradientTab[i]*fx0+gradientTab[i+1]*fy0;
	i=3*IND(ix+1,iy);
	n10=gradientTab[i]*fx1+gradientTab[i+1]*fy0;
	i=3*IND(ix,iy+1);
	n01=gradientTab[i]*fx0+gradientTab[i+1]*fy1;
	i=3*IND(ix+1,iy+1);
	n11=gradientTab[i]*fx1+gradientTab[i+1]*fy1;
	nx0=LERP(wx,n00,n10);	
	nx1=LERP(wx,n01,n11);	
	return LERP(wy,nx0,nx1);
}

void PerFrameInit(float t, float amp, float dx, float scale, float wdir[4]) {
	int lowNoiseRes=8;
	int maxOrder;
	GLfloat *controls;  // control points to add low frequency noise
	GLfloat *deriveU;   // control points for derivatives
	GLfloat *deriveV;   // control points for derivatives
	GLfloat grid[]={MESHRES,MESHRES,MESHRES,MESHRES};
	GLfloat dNoise;
	GLfloat fx,fy;
	int j,k,ju,jv,x,y;
    
	controls=new GLfloat[lowNoiseRes*lowNoiseRes*4];
	deriveU =new GLfloat[(lowNoiseRes-1)*(lowNoiseRes-1)*4];
	deriveV =new GLfloat[(lowNoiseRes-1)*(lowNoiseRes-1)*4];

	float grad[3];
	for (int i=0; i<TABSIZE; i++) {
		CalcGradient(grad,i+0.4f*t);
		LET(&GRADtab[4*i],grad);
	}
	grad[0]=wdir[0]*t+wdir[2];
	grad[1]=wdir[1]*t+wdir[3];
	grad[3]=FLOOR(grad[0]/TABSIZE);
	grad[0]-=grad[3]*TABSIZE;
	grad[3]=FLOOR(grad[1]/TABSIZE);
	grad[1]-=grad[3]*TABSIZE;
	glProgramParameters4fvNV(GL_VERTEX_PROGRAM_NV,31,TABSIZE,(const GLfloat*)GRADtab);						
	glProgramParameter4fvNV(GL_VERTEX_PROGRAM_NV,95,(const GLfloat *)GRADtab);
	glProgramParameter4fNV(GL_VERTEX_PROGRAM_NV,4, grad[0],grad[1],t,scale);
	glProgramParameter4fNV(GL_VERTEX_PROGRAM_NV,5, TABSIZE,1.0f/TABSIZE, 0, 1);
	glProgramParameter4fNV(GL_VERTEX_PROGRAM_NV,7, dx, amp, dx*scale,0);			
	glProgramParameter4fNV(GL_VERTEX_PROGRAM_NV,25, amp, amp, amp,0);			

	// Set up surface evaluator		
	glGetIntegerv(GL_MAX_MAP_TESSELLATION_NV,&maxOrder);	
	if (lowNoiseRes>maxOrder) lowNoiseRes=maxOrder;	  	
	grad[0]=wdir[0]*0.5f*t+0.5f*wdir[2];
	grad[1]=wdir[1]*0.5f*t+0.5f*wdir[3];
	dNoise=(float)(MESHMAX-MESHMIN)/(float)(lowNoiseRes-1);					
	for (y=0; y<lowNoiseRes; y++) {
		fy=MESHMIN+dNoise*(float)y;
		for (x=0; x<lowNoiseRes; x++) {	
			fx=MESHMIN+dNoise*(float)x;
			j=(x+y*lowNoiseRes)*4;
			controls[j  ]=fx;
			controls[j+1]=NORMNULL+5.27*amp*lookupHeight(0.25*scale*fx-grad[0],0.25*scale*fy-grad[1]);
			controls[j+2]=fy;
			controls[j+3]=1.0f;			
		}
	}

	for (y=0; y<lowNoiseRes-1; y++) {		
		for (x=0; x<lowNoiseRes-1; x++) {										
			j=(x+y*lowNoiseRes)*4;
			k=(x+y*(lowNoiseRes-1))*4;
			ju=(x+1+y*lowNoiseRes)*4;
			jv=(x+(y+1)*lowNoiseRes)*4;			
			deriveU[k  ]=(lowNoiseRes)*dNoise; // constant due to regular gridding
			deriveU[k+1]=(lowNoiseRes)*(controls[ju+1]-controls[j+1]);
			deriveU[k+2]=0;        // derivative along x is 0 in z-component
			deriveU[k+3]=1.0f;
			
			deriveV[k  ]=0;
			deriveV[k+1]=(lowNoiseRes)*(controls[jv+1]-controls[j+1]);
			deriveV[k+2]=(lowNoiseRes)*dNoise;
			deriveV[k+3]=1.0f;					
		}		
	}	
	glMapControlPointsNV(GL_EVAL_2D_NV,2,GL_FLOAT,4*sizeof(GL_FLOAT),(lowNoiseRes-1)*4*sizeof(GL_FLOAT),lowNoiseRes-1,lowNoiseRes-1,GL_FALSE,deriveV);
	glMapParameterfvNV(GL_EVAL_2D_NV,GL_MAP_TESSELLATION_NV,grid);
	glDisable(GL_EVAL_FRACTIONAL_TESSELLATION_NV);
	glEnable(GL_EVAL_VERTEX_ATTRIB2_NV);	

	glMapControlPointsNV(GL_EVAL_2D_NV,1,GL_FLOAT,4*sizeof(GL_FLOAT),(lowNoiseRes-1)*4*sizeof(GL_FLOAT),lowNoiseRes-1,lowNoiseRes-1,GL_FALSE,deriveU);
	glMapParameterfvNV(GL_EVAL_2D_NV,GL_MAP_TESSELLATION_NV,grid);
	glDisable(GL_EVAL_FRACTIONAL_TESSELLATION_NV);
	glEnable(GL_EVAL_VERTEX_ATTRIB1_NV);	
	
	glMapControlPointsNV(GL_EVAL_2D_NV,0,GL_FLOAT,4*sizeof(GL_FLOAT),lowNoiseRes*4*sizeof(GL_FLOAT),lowNoiseRes,lowNoiseRes,GL_FALSE,controls);
	glMapParameterfvNV(GL_EVAL_2D_NV,GL_MAP_TESSELLATION_NV,grid);
	glDisable(GL_EVAL_FRACTIONAL_TESSELLATION_NV);
	glEnable(GL_EVAL_VERTEX_ATTRIB0_NV);

	delete[] controls;
	delete[] deriveU;
	delete[] deriveV;
}

/*
 * INITIALIZATION of noise-functions
 * 
 */
void initNoiseH(int seed) {
    gradientTabInit(seed);	
	permutationTabInit(seed+1);
}
#endif