/*
 * SHADER.H - a vertex shader including init-routines
 *
 */

#ifndef __SHADER_H__
#define __SHADER_H__

#include "noise.h"		 // making noise

/*
 * INPUT TO THE VERTEX-SHADER:
 *
 * c[0] : M         Modelview, identity
 * c[4] : Constants (wdir[0]*wspeed*t, wdir[1]*wspeed*t, wspeed*t, scale) [VSP]
 * c[5] : Constants (TABSIZE, 1/TABSIZE, 0, 1) [VSP]
 * c[6] : Constants (0,1,2,3)
 * c[7] : Constants (ds, amplitude, dS, 0)
 * c[8] : MP        Modelview_Projection, identity
 * c[12]: M^(-T)    Modelview, inverse_transpose
 * c[16]: Lc        Lighting coefficients (Ka, Kd, Kt, Nt)
 * c[17]: Ld        Lightdir in OBJECTspace
 * c[18]: C0        Constants (f0, 1.0f, 0.5f, 5.0f)
 * c[19]: C1        Constants (1/snell, NA, NA, 2.0)
 * c[21]: Fc        Fresnel coeffs: (1-f0,1,f0,5.0) [VSP]
 * c[22]: fitCube   (fitX, fitY, fitZ, -0.05f)
 * c[23]: Constants (1, -1, 99.0f, 0.01f)
 * c[24]: Constants (Rabsorp,Gabsorp,Babsorp,NA)
 * c[25]: Constants (amplitude,amplitude,amplitude,0)
 * c[26]-c[30]: NA
 * c[31] - c[95]: Noise tables. xyz -> Gradient, w -> Permutation
 */

#define LOG_BASE2_E 1.442695f

static GLfloat lightspecs[4] ={0.20f, 0.50f,  0.90f, 120.00f};
static GLfloat lightdir[4]   ={0.0f,  -1.0f,  0.0f,  0.0f};
static GLfloat Constants0[4] ={0.20f, 1.00f,  0.50f,  5.00f};
static GLfloat Constants1[4] ={1/snell_coeff, 1.00f,  0.00f,  2.00f};
static GLfloat fitCubeXYZ[4] ={2.0f/(MESHMAX-MESHMIN),2.0f/(MESHMAX-MESHMIN),2.0f/(MESHMAX-MESHMIN),-0.05f};
static GLfloat Constants2[4] ={1,-1,99.0f,0.01f};

static GLuint vertexprog[2],vertexstateprog;
static int workingProg=1;
static int stateProgLoaded=0;

const unsigned char stateProg[]=
{
"!!VSP1.0\
MOV R0.xyzw,c[18].zyxw;\
ADD R0.x,R0.y,-R0.z;\
MOV c[21],R0;\
END"
};

/*
 * Vertex program is in he file "shader.txt" and can be re-loaded interactively.
 *
 */

// Calculate the Fresnel-Coefficient at incident angle
void CalcFresnel_f0(float refraction_coeff, float Gamma, float reflectivity, float absorptions[3]) {
	float constant[4]={0,0,0,2};
	float r=sqrt(refraction_coeff*refraction_coeff+Gamma*Gamma);	
	// calculate f0
	Constants0[0]=(1-1/r)*(1-1/r)/(1+1/r)*(1+1/r)*reflectivity;		
	// calculate 1/snell
	Constants1[0]=1.0f/refraction_coeff;				
	glProgramParameter4fvNV(GL_VERTEX_PROGRAM_NV, 18, Constants0);	
	glProgramParameter4fvNV(GL_VERTEX_PROGRAM_NV, 19, Constants1);		
	constant[0]=LOG_BASE2_E*absorptions[0];
	constant[1]=LOG_BASE2_E*absorptions[1];
	constant[2]=LOG_BASE2_E*absorptions[2];
	glProgramParameter4fvNV(GL_VERTEX_PROGRAM_NV, 24, constant);		
}

// Load the vertex program and track vital matrices.	
int LoadVertexProgram(float snell_coeff) {	
	FILE *fp=NULL;
	unsigned char *vprog;
	long p=0;
	GLint error,line=1;	
	if (!stateProgLoaded) {
		glGenProgramsNV(1,&vertexstateprog);
		glBindProgramNV(GL_VERTEX_STATE_PROGRAM_NV,vertexstateprog);	
		glGetError(); // Reset Error flags.
		glLoadProgramNV (GL_VERTEX_STATE_PROGRAM_NV, vertexstateprog, strlen((const char *)stateProg), (const GLubyte *)stateProg);	
		if (glGetError()!=GL_NO_ERROR) {
			printf("error loading vertex state prog.\n"); 
			exit(0);
		}
		printf("Vertex State Program successfully loaded.\n");
		stateProgLoaded=1;
	}
	fp=fopen("shader.txt","r");
	if (fp==NULL) {
		printf("Vertex Program not found.\n");
		return 0;
	}
	vprog=new unsigned char[64000]; // should be enough.	
	do {
		vprog[p]=(unsigned char)fgetc(fp);
		p++;
	} while (!feof(fp));
	p--;
	fclose(fp);
	
	workingProg=1-workingProg; // switch programs

	glDisable(GL_VERTEX_PROGRAM_NV);
	// Try to bind program
	if (glIsProgramNV(vertexprog[workingProg])) {
		glDeleteProgramsNV(1,&vertexprog[workingProg]);			
	}	
	glGenProgramsNV(1,&vertexprog[workingProg]);
	glBindProgramNV(GL_VERTEX_PROGRAM_NV,vertexprog[workingProg]);		
	glGetError(); // Reset Error flags.
	glLoadProgramNV (GL_VERTEX_PROGRAM_NV, vertexprog[workingProg], p, (const GLubyte *)vprog);	
	if (glGetError()!=GL_NO_ERROR) {
		glGetIntegerv(GL_PROGRAM_ERROR_POSITION_NV,&error);
		// Convert the position into a line-number:
		p=0;
		do {
			if (vprog[p]=='\n') line++;
			p++;
		} while (p<error);
		delete[] vprog;
		printf("Vertex Program: syntax error in line %i.",line);				
		workingProg=1-workingProg; // switch programs
		glBindProgramNV(GL_VERTEX_PROGRAM_NV,vertexprog[workingProg]);			
		glLoadProgramNV (GL_VERTEX_PROGRAM_NV, vertexprog[workingProg], p, (const GLubyte *)vprog);		
		printf(" Old program restored.\n",error);								
		return 0;
	}	
	glTrackMatrixNV(GL_VERTEX_PROGRAM_NV, 0, GL_MODELVIEW, GL_IDENTITY_NV);		
	glTrackMatrixNV(GL_VERTEX_PROGRAM_NV, 8, GL_MODELVIEW_PROJECTION_NV, GL_IDENTITY_NV);
	glTrackMatrixNV(GL_VERTEX_PROGRAM_NV,12, GL_MODELVIEW, GL_INVERSE_TRANSPOSE_NV);			
	glProgramParameter4fNV(GL_VERTEX_PROGRAM_NV, 6, 0,1,2,3);		
	glProgramParameter4fvNV(GL_VERTEX_PROGRAM_NV, 16, lightspecs);
	glProgramParameter4fNV(GL_VERTEX_PROGRAM_NV, 17, 0,-1,0,0);//lightdir);
	glProgramParameter4fvNV(GL_VERTEX_PROGRAM_NV, 18, Constants0);
	glProgramParameter4fvNV(GL_VERTEX_PROGRAM_NV, 19, Constants1);
	glProgramParameter4fvNV(GL_VERTEX_PROGRAM_NV, 22, fitCubeXYZ);
	glProgramParameter4fvNV(GL_VERTEX_PROGRAM_NV, 23, Constants2);

	printf("Vertex Program successfully loaded.\n");
	delete[] vprog;
	return 1;
}

#endif