/*****************************************************************************
 *                                                                           *
 * WATER.CPP                                                                 *
 *                                                                           *
 * authors: Jens Schneider and Ruediger Westermann, (c)2001                  *
 * mails  : jens@glHint.de, westermann@sc.rwth-aachen.de                     *
 * version: 1.1                                                              *
 *                                                                           *
 * This code is under the GNU Public License, for more information see       *
 * the file License.txt                                                      *
 *                                                                           *
 * Sample-code demonstrating real-time rendering of water-ripples.           *
 * Note that the code is highly optimized to show what performance can be    *
 * expected of this technique in future game-releases ;). Sorry if the code  *
 * has become quite un-readable, I've commented it as much as possible.      *
 *                                                                           *
 * To run, you'll need the following extensions to the OpenGL standard:      *
 *                                                                           *
 *    - GL_EXT_texture_cube_map                                              *
 *    - GL_ARB_multitexture                                                  *
 *    - GL_NV_vertex_program                                                 *
 *    - GL_NV_evaluators                                                     *
 *    - GL_EXT_texture_combine_env                                           *
 *                                                                           *
 * The first four extensions are crucial and can be found on nVidia's new    *
 * GeForce3 card. We make extensive use of the GL_NV_vertex_program, so you  *
 * should be familiar with that issue. If you feel unsure about "vertex      *
 * programs" or "vertex shaders", refer to nVidia's documentation to the     *
 * GL_NV_vertex_program.                                                     *
 *                                                                           *
 * For more information see our paper:                                       *
 *                                                                           *
 * J. Schneider and R. Westermann,                                           *
 * "Towards Real-Time Visual Simulation of Water-Surface",                   *
 * in Proceedings 6th Conference on Vision, Modelling and Visualization,     *
 * November 2001, Stuttgart, Germany.                                        *
 *                                                                           *
 * A .pdf version can be downloaded at:                                      *
 *                                                                           *
 * www.vis.rwth-aachen.de [go to "Research"]                                 *
 *                                                                           *
 *****************************************************************************/
#include <math.h>
#include <stdio.h>
#include <stdlib.h>		
#include <GL/glut.h>
#include <string.h>
#include "./extensions/ARB_multitexture.h"
#include "./extensions/ARB_texture_cube_map.h"
#include "./extensions/NV_vertex_program.h"
#include "./extensions/NV_evaluators.h"
#include "./extensions/EXT_texture_env_combine.h"
#include "bmp.h"         // lib containing a 24bit-bmp loader. From nehe.gamedev.net tutorials.
#include "modells.h"     // contains a vertex-array for the pool-frame
#include "mymath.h"      // fast sine/cosine tables.


#define GL_TEXTURE_WRAP_R 0x8072 // I guess nVidia forgot this in new drivers

/*
 *  YOU MAY MODIFY THE FOLLOWING PARAMETERS:
 *
 */
#define MESHRES 128              // geometric resolution of the mesh
static float snell_coeff=1.33f; // optical index for water
static float extinct_coeff[3]={0.5f,0.4f,0.3f};// RGB-extinction coefficient for water.

// See shader.h for more tweakable items.

/*
 * CODE FOR TIMINGS (Linux & Win9x)
 *
 */
#ifdef _WIN32
	#include <sys/timeb.h>

	#  define TIMER_VAR(x)  struct _timeb ts_ ## x, te_ ## x
	#  define TIMER_START(x) _ftime (&ts_ ## x)
	#  define TIMER_PROBE(x)  _ftime (&te_ ## x)
	#  define TIMER_RESULT(x) ((double)te_ ## x.time - (double)ts_ ## x.time + ((double)te_ ## x.millitm - (double)ts_ ## x.millitm)/1e3)

#else

	#  include <unistd.h>
	#  include <sys/time.h>
	#  include <sys/times.h>
	#  include <sys/types.h>

	#  define TIMER_VAR(x)   struct timeval ts_ ## x , te_ ## x
	#  define TIMER_START(x) gettimeofday (&ts_ ## x, 0)
	#  define TIMER_PROBE(x)  gettimeofday (&te_ ## x, 0)
	#  define TIMER_RESULT(x) ((double)te_ ## x.tv_sec - (double)ts_ ## x.tv_sec + ((double)te_ ## x.tv_usec - (double)ts_ ## x.tv_usec)/1e6)
#endif

/*
 * INTERNAL DEFs. Rather don't touch them.
 *
 */ 
#define MESHMIN (-2.0f)
#define MESHMAX 2.0f
#define MESHDIM (MESHRES+1)
#define INDEX(i,j) ((i)*MESHDIM+(j))
#define NORMNULL -0.5f
static float   ds = (MESHMAX-MESHMIN)/(MESHRES-1);	


/*
 * MORE INCLUDES
 *
 */
#include "shader.h"
#include "noise.h"		 // making noise


/*
 * SYSTEM VARs. Rather don't touch them.
 *
 */

static GLsizei winwidth = 512;
static GLsizei winheight= 512;
static float   aspect = 1.0f;
static float   rotx   = 0.0f;
static float   roty   = 45.0f;
static int     lflag  = 0; 
static int     rflag  = 0;
static int     mx,my;
static float   wn     = 0.5f;
static float   tz     = 0;
static float   speed  = 3.0f;
static float   windRot= 45.0f;
static float   amplitude = 0.5f;
static int     toggle_wireframe  =0;
static int     toggle_flatsurface=0;
static float   reflectivity = 1.0f;
static float   WindDir[4]={0,0,0,0};

/*
 * LIGHTSOURCE.
 *
 */
static GLfloat ambient_color[4]  = {+0.40f, +0.40f, +0.40f, +1.0f};
static GLfloat diffuse_color[4]  = {+0.50f, +0.50f, +0.50f, +1.0f};
static GLfloat specular_color[4] = {+0.90f, +0.90f, +0.90f, +1.0f};
static GLfloat white[4]          = {+1.00f, +1.00f, +1.00f, +1.0f};
static GLfloat light_dir[4]      = { 0.00f, -4.00f, +0.20f, +0.0f};
static GLfloat Nt                = 24.0f;

/*
 * OPENGL-HANDLERs for textures and display lists.
 *
 */
static GLuint watertex,cubetex;
static GLuint poolframe,Arrows;

/*
 * VERTEX ARRAYS
 *
 */
unsigned long int BUFFERSIZE=MESHDIM*MESHDIM*sizeof(GLfloat);

/*
 * FOR TIMING
 *
 */
TIMER_VAR(timevar);
static double   elapsed_time;
static double   fps = 0.0;
static double   last;
static double   lastfps;

/*
 * LOAD AND BIND TEXTURES
 *
 */
void LoadTextures() {
    textureImage image;	
    
    // Load and bind Cubemap-Texture		
    glGenTextures(1,&cubetex);
    glBindTexture(GL_TEXTURE_CUBE_MAP_ARB,cubetex);
    glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
    glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB,GL_TEXTURE_MAG_FILTER,GL_LINEAR);		
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB,GL_TEXTURE_WRAP_T,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB,GL_TEXTURE_WRAP_R,GL_CLAMP);
    LoadBmp("pos_X.bmp",&image);
    gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB, GL_RGB8,image.width,image.height, GL_RGB, GL_UNSIGNED_BYTE, image.data);   	
    LoadBmp("neg_X.bmp",&image);
    gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB, GL_RGB8,image.width,image.height, GL_RGB, GL_UNSIGNED_BYTE, image.data);   	
    LoadBmp("pos_Y.bmp",&image);
    gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB, GL_RGB8,image.width,image.height, GL_RGB, GL_UNSIGNED_BYTE, image.data);
    LoadBmp("neg_Y.bmp",&image);
    gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB, GL_RGB8,image.width,image.height, GL_RGB, GL_UNSIGNED_BYTE, image.data);
    LoadBmp("pos_Z.bmp",&image);
    gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB, GL_RGB8,image.width,image.height, GL_RGB, GL_UNSIGNED_BYTE, image.data);
    LoadBmp("neg_Z.bmp",&image);
    gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB, GL_RGB8,image.width,image.height, GL_RGB, GL_UNSIGNED_BYTE, image.data);
    if (image.data) free(image.data);	
}

/*
 * GENERATE DISPLAY LISTS for the pool-frame and the arrow
 *
 */
void initmodells(void) {
    int i;
    poolframe=glGenLists(1);
    glNewList(poolframe,GL_COMPILE);		
        glBegin(GL_QUADS);
            for (i=0; i<frame_entries; i++) {
                glMultiTexCoord3fvARB(GL_TEXTURE0_ARB,&frame[9*i]);
                glNormal3fv(&frame[9*i+3]);
                glVertex3fv(&frame[9*i+6]);
            }
        glEnd();
    glEndList();

    Arrows=glGenLists(1);
    glNewList(Arrows,GL_COMPILE);
    drawArrow();
    glEndList();
}		

/*
 * GENERAL initialization and setup
 *
 */
void init(void) {	
    float dir[4];	
	GLfloat controls[]={ MESHMIN,NORMNULL,MESHMIN,1,
		                 MESHMAX,NORMNULL,MESHMIN,1,
					     MESHMIN,NORMNULL,MESHMAX,1,
					     MESHMAX,NORMNULL,MESHMAX,1};						 
	GLfloat grid[]={MESHRES,MESHRES,MESHRES,MESHRES};
    // normalize lightdirection:
    normalize(light_dir);
	
    // OpenGL needs -light_dir
    dir[0]=-light_dir[0];
    dir[1]=-light_dir[1];
    dir[2]=-light_dir[2];
    dir[3]=-light_dir[3];

    // Set up OpenGL-Lighting for poolframe etc.
    glLightfv(GL_LIGHT0,GL_AMBIENT,ambient_color);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse_color);
    glLightfv(GL_LIGHT0, GL_SPECULAR, specular_color);
    glLightfv(GL_LIGHT0, GL_POSITION, dir);	
    glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER,1.0f);
    glEnable(GL_LIGHT0);
    glDisable(GL_LIGHTING);		
    glMaterialf(GL_BACK,GL_SHININESS,Nt);
    glMaterialfv(GL_BACK,GL_SPECULAR,white);	
    glMaterialfv(GL_BACK,GL_DIFFUSE,white);	
    glMaterialfv(GL_BACK,GL_AMBIENT,white);		

	// Test and bind extensions
	if (!GL_ARB_multitexture_Init()) exit(0);
	if (!GL_ARB_texture_cube_map_Init()) exit(0);
	if (!GL_NV_vertex_program_Init()) exit(0);
	if (!GL_EXT_texture_env_combine_Init()) exit(0);	
	if (!GL_NV_evaluators_Init()) exit(0);	
	if (!LoadVertexProgram(snell_coeff)) exit(0);

	// initialize tables and arrays, load textures
	initNoiseH(10);	
	CalcFresnel_f0(snell_coeff,extinct_coeff[1],reflectivity,extinct_coeff); // use G extinction coeff.		
	initmodells();
	LoadTextures();	

    // OpenGL setup, Texture-Coordinate generation etc.
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);		
    glShadeModel(GL_SMOOTH);	
    glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
    glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);
    glCullFace(GL_BACK);    

/*
 *  BLENDING equation: 
 *  T0: texture0, refraction map
 *  T1: texture1, reflection map
 *  C0: primary color, .rgb: lighting coded in RGB components, .a : reflectivity coded in Alpha-component
 *  
 *  Cp= (1 - C0.a) * (C0.rgb * T0) + C0.a * T1
 */

    glActiveTextureARB(GL_TEXTURE0_ARB);

    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);
    glTexEnvi(GL_TEXTURE_ENV, (GLenum)GL_COMBINE_RGB_EXT, GL_MODULATE);
    glTexEnvi(GL_TEXTURE_ENV, (GLenum)GL_SOURCE0_RGB_EXT, GL_PRIMARY_COLOR_EXT);
    glTexEnvi(GL_TEXTURE_ENV, (GLenum)GL_OPERAND0_RGB_EXT, GL_SRC_COLOR);
    glTexEnvi(GL_TEXTURE_ENV, (GLenum)GL_SOURCE1_RGB_EXT, GL_TEXTURE);
    glTexEnvi(GL_TEXTURE_ENV, (GLenum)GL_OPERAND1_RGB_EXT, GL_SRC_COLOR);	
    glBindTexture(GL_TEXTURE_CUBE_MAP_ARB,cubetex);

	glDisable(GL_TEXTURE_2D);
    glEnable((GLenum)GL_TEXTURE_CUBE_MAP_ARB);

    glActiveTextureARB(GL_TEXTURE1_ARB);	

    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);
    glTexEnvi(GL_TEXTURE_ENV, (GLenum)GL_COMBINE_RGB_EXT, GL_INTERPOLATE_EXT);			
    glTexEnvi(GL_TEXTURE_ENV, (GLenum)GL_SOURCE0_RGB_EXT, GL_TEXTURE);		
    glTexEnvi(GL_TEXTURE_ENV, (GLenum)GL_OPERAND0_RGB_EXT, GL_SRC_COLOR);
    glTexEnvi(GL_TEXTURE_ENV, (GLenum)GL_SOURCE1_RGB_EXT, GL_PREVIOUS_EXT);
    glTexEnvi(GL_TEXTURE_ENV, (GLenum)GL_OPERAND1_RGB_EXT, GL_SRC_COLOR);
    glTexEnvi(GL_TEXTURE_ENV, (GLenum)GL_SOURCE2_RGB_EXT, GL_PRIMARY_COLOR_EXT);
    glTexEnvi(GL_TEXTURE_ENV, (GLenum)GL_OPERAND2_RGB_EXT, GL_SRC_ALPHA);
    glBindTexture(GL_TEXTURE_CUBE_MAP_ARB,cubetex);
    
    glTexGeni(GL_S,GL_TEXTURE_GEN_MODE,GL_REFLECTION_MAP_ARB);
    glTexGeni(GL_T,GL_TEXTURE_GEN_MODE,GL_REFLECTION_MAP_ARB);	
    glTexGeni(GL_R,GL_TEXTURE_GEN_MODE,GL_REFLECTION_MAP_ARB);	
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);	
    glEnable(GL_TEXTURE_GEN_R);	
	glDisable(GL_TEXTURE_2D);
    glEnable((GLenum)GL_TEXTURE_CUBE_MAP_ARB);	

    TIMER_START(timevar);
	
	// Set up surface evaluator	
	glMapControlPointsNV(GL_EVAL_2D_NV,0,GL_FLOAT,4*sizeof(GL_FLOAT),8*sizeof(GL_FLOAT),2,2,GL_FALSE,controls);
	glMapParameterfvNV(GL_EVAL_2D_NV,GL_MAP_TESSELLATION_NV,grid);
	glDisable(GL_EVAL_FRACTIONAL_TESSELLATION_NV);
	glEnable(GL_EVAL_VERTEX_ATTRIB0_NV);		
    
	WindDir[0]=DCOS(windRot);
	WindDir[1]=-DSIN(windRot);	        	    
}

/*
 * Per Frame-Setup performed here.
 *
 */
void CalcPerturbation(void) {
    float t;    
    
    t=speed*elapsed_time;
 
    // (float t, float amp, float dx, float scale, float wdir[2])
	PerFrameInit(0.5f*t,0.15f*amplitude,0.1f*ds,4.0f*wn,WindDir);	// time-animation. the factor 0.5 can be treated as wind-speed.	
}

void displayfunc(void) {
	GLfloat params[4];
	TIMER_PROBE(timevar);
	elapsed_time=TIMER_RESULT(timevar);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();
	glTranslatef(0,+0.2,-4-tz);
	glRotatef(rotx,0,1,0);
	glRotatef(roty,1,0,0);				
		
    CalcPerturbation();	

	glActiveTextureARB(GL_TEXTURE1_ARB);	
	
	glExecuteProgramNV(GL_VERTEX_STATE_PROGRAM_NV,vertexstateprog,params);
	glEnable(GL_VERTEX_PROGRAM_NV);
	glEnable(GL_CULL_FACE);
    glEdgeFlag(1);

	// Evaluate the surface
	glEvalMapsNV(GL_EVAL_2D_NV,GL_FILL);	

	glDisable(GL_VERTEX_PROGRAM_NV);
    
	// Cubemapping has higher priority than 2D mapping. 
	// This is exploited to reduce state changes.
	glDisable((GLenum)GL_TEXTURE_CUBE_MAP_ARB);	
	glDisable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
	glColor4f(1,1,1,0.1);	
	
	glCallList(poolframe);

	glRotatef(windRot,0,1,0);
	glTranslatef(-2.0f,0.5f,0.0f);

    glActiveTextureARB(GL_TEXTURE0_ARB);    
    glDisable((GLenum)GL_TEXTURE_CUBE_MAP_ARB);

    glColor4f(1,0,0,0.1f);
    glCallList(Arrows);
	
    glEnable((GLenum)GL_TEXTURE_CUBE_MAP_ARB);
    glActiveTextureARB(GL_TEXTURE1_ARB);    
	glEnable((GLenum)GL_TEXTURE_CUBE_MAP_ARB);
    
	glDisable(GL_LIGHTING);
    
    fps+=1.0f;		
    glutSwapBuffers();

}

void reshapefunc(int width,int height)
{
	winwidth=width;
	winheight=height;
	aspect=(float)width/(float)height;

	glViewport(0,0,width,height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0,aspect,0.2,10.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();	
	glutPostRedisplay();
}


void mousefunc(int button,int state,int x,int y) {
	if (button==GLUT_LEFT_BUTTON) {
		if (state==GLUT_DOWN) lflag=1;
		else lflag=0;
	}	
	else if (button==GLUT_RIGHT_BUTTON) {
		if (state==GLUT_DOWN) rflag=1;
		else rflag=0;
	}
}


void motionfunc(int x,int y) {
	if (lflag>0) {
		if (lflag>1) {
			rotx+=360.0f*(x-mx)/winwidth;
			roty+=360.0f*(y-my)/winheight;
		}
		mx=x;
		my=y;
		lflag=2;
	}
	else if (rflag>1) {
		tz-=4.0*(y-my)/winheight;		
   	}		
	mx=x;
	my=y;
	rflag=2;
	glutPostRedisplay();		
}

void printHelp(void) {
	printf("\n");
	printf("\nTowards Real-Time Visual Simulation of Water-Surfaces\n");
	printf("  (c)2001 by Jens Schneider and Ruediger Westermann\n");
	printf("=======================================================\n");
	printf(" Use mouse to move the pool.\n");
	printf(" h,H this help text.\n");
	printf(" w,W toggle wireframe mode. \n");
	printf(" r,R increases/decreases reflectivity.\n");
	printf(" d,D rotate wind-direction.\n");
	printf(" f,F increase/decrease frequencies.\n");
	printf(" x,X increase/decrease wavenumbers.\n");
	printf(" a,A increase/decrease aplitudes.\n");
	printf(" +,- increase/decrease optical density.\n");	
	printf(" s,S toggle flat surface.\n");
	printf(" v,V reload Vertex program.\n");	
}
	
	   
void keyboardfunc(unsigned char key,int x,int y) {	
	float dir[2];
	float t=0.5f*speed*elapsed_time;
 	switch(key) {
	case 'V':
	case 'v':
		LoadVertexProgram(snell_coeff);
		break;
	case 'R':
		reflectivity-=0.05f;
		if (reflectivity<0.0f) reflectivity=0.0f;
		CalcFresnel_f0(snell_coeff,extinct_coeff[1],reflectivity,extinct_coeff);		
		break;	
	case 'r':
		reflectivity+=0.05f;
		if (reflectivity>5.0f) reflectivity=5.0f;
		CalcFresnel_f0(snell_coeff,extinct_coeff[1],reflectivity,extinct_coeff);		
		break;		
	case 'h':
	case 'H':
		printHelp();
		break;
    case 'd':
        windRot+=3.0f;
		dir[0]=DCOS(windRot);
		dir[1]=-DSIN(windRot);	        
		WindDir[2]=(WindDir[0]-dir[0])*t+WindDir[2];
		WindDir[3]=(WindDir[1]-dir[1])*t+WindDir[3];
		WindDir[0]=dir[0];
		WindDir[1]=dir[1];
        break;
    case 'D':
        windRot-=3.0f; 
		dir[0]=DCOS(windRot);
		dir[1]=-DSIN(windRot);	        
		WindDir[2]=(WindDir[0]-dir[0])*t+WindDir[2];
		WindDir[3]=(WindDir[1]-dir[1])*t+WindDir[3];
		WindDir[0]=dir[0];
		WindDir[1]=dir[1];
        break;
	case 'f':
		speed+=0.05f;
		if (speed>5.0f) speed=5.0f;				
		break;
	case 'F':
		speed-=0.05f;
		if (speed<0.0f) speed=0.0f;
		break;
	case 'x':		
		wn+=0.025f;
		if (wn>5.0f) wn=5.0f;
		break;
	case 'X':		
		wn-=0.025f;
		if (wn<0.0f) wn=0.0f;
		break;
	case 'a':
        amplitude+=0.1f;
        if (amplitude>5.0f) amplitude=5.0f;
        break;
    case 'A':
        amplitude-=0.1f;
        if (amplitude<-5.0f) amplitude=-5.0f;
        break;
	case 'w':
	case 'W':
		toggle_wireframe=1-toggle_wireframe;
		if (toggle_wireframe) {
			glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
			glDisable(GL_CULL_FACE);
		}
		else {
			glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
			glEnable(GL_CULL_FACE);
		}
		break;
	case '+': 
		snell_coeff+=0.05f; 
		CalcFresnel_f0(snell_coeff,extinct_coeff[1],reflectivity,extinct_coeff);		
		break;
	case '-': 
		snell_coeff-=0.05f; 
		CalcFresnel_f0(snell_coeff,extinct_coeff[1],reflectivity,extinct_coeff);		
		break;
	case 's':
	case 'S':
		toggle_flatsurface=1-toggle_flatsurface;
		break;
    case 'q': 
	case 27:	
		glDeleteTextures(1,&cubetex);
		glDeleteLists(Arrows,1);
		glDeleteLists(poolframe,1);
		exit(0);
		break;
	 default:
		printf("undefined key\n");
	}
	glutPostRedisplay();
}


void specialfunc(int key,int x,int y) {}

/*
 * DISPLAY UPDATED INFORMATION IN THE TITLEBAR
 *
 */ 

// write infos to the title-bar.
void UpdateTitle(void) {
	char text[100];
	sprintf(text,"Visual Simulation of Water-Surfaces (2x%ix%i Tris @ %.3f fps)",MESHRES,MESHRES,lastfps);
	glutSetWindowTitle(text);
}

void idlefunc(void) {
	double probe;
	TIMER_PROBE(timevar);
	probe=TIMER_RESULT(timevar);
	if (probe-last>1.0) {
		lastfps=fps/(probe-last);
		last=probe;
		fps=0.0;
		UpdateTitle();
	}
	glutPostRedisplay();
}


int main(int argc, char **argv) {
	glutInit(&argc,argv);
	glutInitWindowSize(512,512);
	glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA|GLUT_DEPTH);
	glutCreateWindow("Water");
	
	glutDisplayFunc(displayfunc);
	glutReshapeFunc(reshapefunc);
	glutMouseFunc(mousefunc);
	glutMotionFunc(motionfunc);
	glutKeyboardFunc(keyboardfunc);
	glutSpecialFunc(specialfunc);
	glutIdleFunc(idlefunc);
	init();
	
	printHelp();
	
	TIMER_PROBE(timevar);
	last=TIMER_RESULT(timevar);
	UpdateTitle();	

	glutMainLoop();
	return(0);
}
